(import tree-walkers)

(import simple-tests (chicken base))

(define-tester
  (walkers?)
  ((iterate cdr 2) '(1 2 3))
  '(3)
  ((walk -2 1 -1 1) '(a (b (c)) d))
  'c
  ((walk 3) '(a (b (c)) d))
  '()
  ((walk -4) '((((x)))))
  'x
  ((walk -1 1 -1 1) '(a (b x) c))
  'x
  ((walk -1 1 -1 1) '(a (b x) c y))
  'x
  ((walk -1 1 1 1) '(a (b x) c y))
  'y)

(define-tester
  (pathes?)
  (tree->pathes '(a b . c))
  '((a (-1)) (b (-1 1)) (c (1 1)))
  (tree->pathes '(a b c))
  '((a (-1)) (b (-1 1)) (c (-1 1 1)) (() (1 1 1)))
  (tree->pathes '(a (b c) d))
  '((a (-1))
    (d (-1 1 1))
    (() (1 1 1))
    (b (-1 -1 1))
    (c (-1 1 -1 1))
    (() (1 1 -1 1)))
  (tree->pathes '(a (b (c)) d))
  '((a (-1))
    (d (-1 1 1))
    (() (1 1 1))
    (b (-1 -1 1))
    (() (1 1 -1 1))
    (c (-1 -1 1 -1 1))
    (() (1 -1 1 -1 1)))
  (tree->pathes '(a (b . c) . d))
  '((a (-1)) (d (1 1)) (b (-1 -1 1)) (c (1 -1 1)))
  (pathes 'a '(((((a))))))
  '((-1 -1 -1 -1 -1))
  (pathes 'c '(a (b . c) d))
  '((1 -1 1))
  (pathes '() '(a (b (c)) d))
  '((1 1 1) (1 1 -1 1) (1 -1 1 -1 1))
  (pathes 'c '(a (b (c)) d))
  '((-1 -1 1 -1 1))
  (pathes 'a '(a a))
  '((-1) (-1 1))
  (pathes 'a 'a)
  '(())
  (map (lambda (triple) `(,(car triple) ,(caddr triple)))
       (tree->accessors '(a (b . c) d) 'walk))
  '((a (walk -1))
    (d (walk -1 1 1))
    (() (walk 1 1 1))
    (b (walk -1 -1 1))
    (c (walk 1 -1 1))))

(define-tester
  (accessors?)
  (let ((pat '(a (b . c) . d)))
    (map (lambda (proc) (proc pat)) (map cadr (tree->accessors pat))))
  '(a d b c)
  (let ((pat '(a (b c) d)))
    (map (lambda (proc) (proc pat)) (map cadr (tree->accessors pat))))
  '(a d () b c ())
  (let ((pat '(a (b (c . d)) (e . f))))
    (map (lambda (proc) (proc pat)) (map cadr (tree->accessors pat))))
  '(a () b () c d e f))

(define-tester
  (predicates?)
  ((pseudo-list-of? atom?) '(a b c))
  #t
  ((pseudo-list-of? symbol?) '(a b c))
  #f
  ((pseudo-list-of? symbol?) '(a b . c))
  #t
  ((pseudo-list-of? atom?) '())
  #t
  ((pseudo-list-of? symbol?) 'a)
  #t
  ((tree-of? symbol?) '(a (b . c) . d))
  #t
  ((tree-of? symbol?) '(a (b . c) d))
  #f
  ((tree-of? (disjoin symbol? null?)) '(a (b . c) d))
  #t
  ((all-match? '(a (b . c) . d)) '(a (b . c) . d))
  #t
  ((all-match? '(a (b . c) . d)) '(1 (2 3) 4))
  #t
  ((all-match? '(a (b c) d)) '(1 (2 3) 4))
  #t
  ((all-match? '(a (#f c) d)) '(1 (2 3) 4))
  #t
  ((all-match? '(a (#f c) d)) '(1 (#f 3) 4))
  #t
  ((all-match? '(a b)) '(1))
  #f
  (same-shape? '(a (b (c) d) e) '(1 (2 (3) 4) 5))
  #t
  (same-shape? '(a (b (c) d) e) '(1 (2 (3 0) 4) 5))
  #f
  (and? (eqv? 'a 'a) (eqv? 1 1) (equal? "foo" "foo"))
  #t
  (and? (eqv? 'a 'a) (equal? "foo" "bar") (eqv? 1 1))
  #f
  (none? odd? '(0 2))
  #t
  (none? odd? '(0 1 2 3))
  #f
  (some? odd? '(0 2))
  #f
  (some? odd? '(1 2))
  #t
  (every? odd? '(1 3))
  #t
  (every? odd? '(0 1 2 3))
  #f
  ((all? same-shape?) '((1 2) (10 20) (100 200)))
  #t
  ((all? same-shape?) '((1 2) (10 20) (100 200)))
  #t
  ((all? same-shape?) '((1 2) (10 . 20) (100 200)))
  #f)

(define-tester
  (mapping?)
  (map* add1 0)
  1
  (map* add1 '(0 (1 2)))
  '(1 (2 3))
  (map* add1 '(0 (1 . 2)))
  '(1 (2 . 3))
  (when-map* (list-of? atom?) list->vector '((1) (2)))
  '(#(1) #(2))
  (when-map* (list-of? atom?) list->vector '((1) ((2) (3))))
  '(#(1) (#(2) #(3)))
  (when-map* (list-of? atom?) list->vector '((1) ((2) 0 (3))))
  '(#(1) (#(2) 0 #(3)))
  (when-map* (conjoin integer? even?) add1 '(0 (1 2 (3))))
  '(1 (1 3 (3)))
  (list-values map** list '(a b) '(1 2))
  '(((a) (b)) ((1) (2)))
  (list-values map** list '(a (b c) d) '(1 (2 (3 4)) #(5 6)))
  '(((a) ((b) (c)) (d)) ((1) ((2) ((3) (4))) (#(5 6))))
  (list-values map** list '(a b) '(1 2) '(10 20))
  '(((a) (b)) ((1 10) (2 20)))
  (list-values map** list '(a (b c) d) '(1 (2 (3)) 4) '(10 (20 (30)) 40))
  '(((a) ((b) (c)) (d)) ((1 10) ((2 20) ((3 30))) (4 40))))

(define-tester
  (filtering?)
  (values->list (filter odd? '(0 1 2 3)))
  '((1 3) (0 2))
  (list-values filter odd? '(0 1 2 3))
  '((1 3) (0 2))
  (values->list (filter-pairs odd? '((0 a) (1 b) (2 c) (3 d))))
  '(((1 b) (3 d)) ((0 a) (2 c)))
  (list-values filter-pairs odd? '((0 a) (1 b) (2 c) (3 d)))
  '(((1 b) (3 d)) ((0 a) (2 c)))
  (filter* odd? '(0 1 2 3))
  '(1 3)
  (filter* odd? '(0 (1 2 3)))
  '((1 3))
  (filter* odd? '(0 1 (2 3 (4 5) 6 7) 8 9))
  '(1 (3 (5) 7) 9)
  (filter* odd? '(0 (2 (4) 6) 8))
  '((()))
  (filter* odd? '(1 (3 (5) 7) 9))
  '(1 (3 (5) 7) 9))

(define-tester
  (procedures?)
  (values->list (list->values '(1 2 3)))
  '(1 2 3)
  (values->list (list->values (values->list (values 1 2 3))))
  '(1 2 3)
  (values->list (split-at 2 '(0 1 2 3)))
  '((0 1) (2 3))
  (list-values split-at 2 '(0 1 2 3))
  '((0 1) (2 3))
  (drop 2 '(0 1 2 3))
  '(2 3)
  (take 2 '(0 1 2 3))
  '(0 1)
  (values->list (split-when odd? '(0 1 2 3)))
  '((0) (1 2 3))
  (list-values split-when odd? '(0 1 2 3))
  '((0) (1 2 3))
  (take-while odd? '(1 3 2 4))
  '(1 3)
  (drop-while odd? '(1 3 2 4))
  '(2 4)
  (memp integer? '(a b c))
  #f
  (memp integer? '(a b 1 c))
  '(1 c)
  (sublist '(0 1 2 3 4) 1 3)
  '(1 2)
  (sublist '(0 1 2 3 4) 4)
  '(4)
  (indexes '(a b c d))
  '(0 1 2 3)
  (unzip* '(1 2 3))
  '(1 2 3)
  (unzip* '((1 2) (1 2) (1 2)))
  '((1 1 1) (2 2 2))
  (unzip* '((1 . 2) (1 . 2) (1 . 2)))
  '((1 1 1) 2 2 2)
  (unzip* '((1 (2 3)) (1 (2 3)) (1 (2 3))))
  '((1 1 1) ((2 2 2) (3 3 3)))
  (unzip* '((0 (1 . 2)) (0 (1 . 2)) (0 (1 . 2))))
  '((0 0 0) ((1 1 1) 2 2 2))
  (unzip* '((0 (1 2)) (0 (1 2)) (0 (1 2))))
  '((0 0 0) ((1 1 1) (2 2 2)))
  (unzip* '((0 (1 2) . 3) (0 (1 2) . 3) (0 (1 2) . 3)))
  '((0 0 0) ((1 1 1) (2 2 2)) 3 3 3)
  (unzip* '((0 (1 . 2) . 3) (0 (1 . 2) . 3) (0 (1 . 2) . 3)))
  '((0 0 0) ((1 1 1) 2 2 2) 3 3 3)
  (flatten* (map* list '(a b (c d)) '(1 2 (#(3 30) #(4 40)))))
  '((a 1) (b 2) (c #(3 30)) (d #(4 40)))
  (flatten* '((a 1) ((b 2)) (((c 3)))))
  '((a 1) (b 2) (c 3))
  (pseudo-head '(1 2 . 3))
  '(1 2)
  (pseudo-tail '(1 2 . 3))
  3
  (pseudo-head '())
  '()
  (pseudo-tail '())
  '()
  (pseudo-head 3)
  '()
  (pseudo-tail 3)
  3
  (pseudo-length '(0 1 . 2))
  2)

(test-all TREE-WALKERS
  walkers?
  pathes?
  accessors?
  predicates?
  mapping?
  filtering?
  procedures?
  )
